"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
import os
import io
import re
import gemmi
import numpy
import pandas
import string
import random

default_proton_scale = 1.13 # scale of X-proton distance to X-H(e) distance

def decide_new_mod_id(mod_id, mods):
    # Refmac only allows up to 8 characters
    letters = string.digits + string.ascii_lowercase
    if len(mod_id) < 8:
        for l in letters:
            new_id = "{}{}{}".format(mod_id, "" if len(mod_id)==7 else "-", l)
            if new_id not in mods:
                return new_id

    # give up keeping original name
    while True: # XXX risk of infinite loop.. less likely though
        new_id = "mod" + "".join([random.choice(letters) for _ in range(4)])
        if new_id not in mods:
            return new_id
# decide_new_mod_id()

def rename_cif_modification_if_necessary(doc, known_ids):
    # FIXME Problematic if other file refers to modification that is renamed in this function - but how can we know?
    
    mod_list = doc.find_block("mod_list")
    if not mod_list: return {}
    
    trans = {}
    for row in mod_list.find("_chem_mod.", ["id"]):
        mod_id = row.str(0)
        if mod_id in known_ids:
            new_id = decide_new_mod_id(mod_id, known_ids)
            trans[mod_id] = new_id
            row[0] = new_id # modify id
            logger.write("INFO:: renaming modification id {} to {}".format(mod_id, new_id))

    # modify ids in mod_* blocks
    for mod_id in trans:
        b = doc.find_block("mod_{}".format(mod_id))
        if not b: # should raise error?
            logger.write("WARNING:: inconsistent data_mod_list for {} in {}".format(mod_id, f))
            continue
        b.name = "mod_{}".format(trans[mod_id]) # modify name
        for item in b:
            for tag in item.loop.tags:
                if tag.endswith(".mod_id"):
                    for row in b.find(tag[:tag.rindex(".")+1], ["mod_id"]):
                        row[0] = trans[mod_id]

    # Update mod id in links
    link_list = doc.find_block("link_list")
    if trans and link_list:
        for row in link_list.find("_chem_link.", ["mod_id_1", "mod_id_2"]):
            for i in range(2):
                if row.str(i) in trans:
                    row[i] = trans[row.str(i)]

    return trans
# rename_cif_modification_if_necessary()

def load_monomer_library(st, monomer_dir=None, cif_files=None, stop_for_unknowns=False, check_hydrogen=False,
                         ignore_monomer_dir=False):
    resnames = st[0].get_all_residue_names()

    if monomer_dir is None and not ignore_monomer_dir:
        if "CLIBD_MON" not in os.environ:
            logger.error("WARNING: CLIBD_MON is not set")
        else:
            monomer_dir = os.environ["CLIBD_MON"]

    if cif_files is None:
        cif_files = []
        
    if monomer_dir and not ignore_monomer_dir:
        if not os.path.isdir(monomer_dir):
            logger.error("ERROR: not a directory: {}".format(monomer_dir))
            return

        logger.write("Reading monomers from {}".format(monomer_dir))
        monlib = gemmi.read_monomer_lib(monomer_dir, resnames, ignore_missing=True)
    else:
        monlib = gemmi.MonLib()

    for f in cif_files:
        logger.write("Reading monomer: {}".format(f))
        doc = gemmi.cif.read(f)
        comp_list = doc.find_block("comp_list")
        for b in doc:
            if b.find_values("_chem_comp_atom.atom_id"):
                name = b.name.replace("comp_", "")
                if name in monlib.monomers:
                    logger.write("WARNING:: updating monomer {} using {}".format(name, f))
                    del monlib.monomers[name]
                monlib.add_monomer_if_present(b)

                if comp_list:
                    tab = comp_list.find("_chem_comp.", ["id", "group"])
                    if tab:
                        monlib.monomers[name].group = tab.find_row(name).str(1)
                
                # Check if bond length values are included
                # This is to fail if cif file is e.g. from PDB website
                for b in monlib.monomers[name].rt.bonds:
                    if b.value != b.value:
                        raise RuntimeError("{} does not contain bond length value for {}. You need to generate restraints (e.g. using acedrg).".format(f, name))
                    
        link_list = doc.find_block("link_list")
        if link_list:
            for row in link_list.find("_chem_link.", ["id"]):
                link_id = row.str(0)
                if link_id in monlib.links:
                    logger.write("WARNING:: updating link {} using {}".format(link_id, f))
                    del monlib.links[link_id]

        # If modification id is duplicated, need to rename
        rename_cif_modification_if_necessary(doc, monlib.modifications)
        
        monlib.insert_chemlinks(doc)
        monlib.insert_chemmods(doc)
        monlib.insert_comp_list(doc)

    not_loaded = set(resnames).difference(monlib.monomers)
    if not_loaded:
        logger.write("WARNING: monomers not loaded: {}".format(" ".join(not_loaded)))
        
    logger.write("Monomer library loaded: {} monomers, {} links, {} modifications".format(len(monlib.monomers),
                                                                                          len(monlib.links),
                                                                                          len(monlib.modifications)))
    logger.write("       Monomers: {}".format(" ".join([x for x in monlib.monomers])))
    logger.write("          Links: {}".format(" ".join([x for x in monlib.links])))
    logger.write("  Modifications: {}".format(" ".join([x for x in monlib.modifications])))
    logger.write("")

    logger.write("Checking if unknown atoms exist..")
    st = st.clone()
    sio = io.StringIO()
    topo = gemmi.prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.NoChange, warnings=sio, reorder=True,
                                  ignore_unknown_links=True)

    # possible warnings:
    # Warning: unknown chemical component XXX in chain X
    # Warning: no atom X expected in XXX
    # and others?
    unknown_cc = set()
    unknown_atoms_cc = set()
    link_related = set()
    for l in sio.getvalue().splitlines():
        r = re.search("Warning: unknown chemical component (.*) in chain", l)
        if r:
            unk = r.group(1)
            if unk not in unknown_cc:
                logger.write(l)
                unknown_cc.add(unk)
            continue
        r = re.search("Warning: definition not found for [^/]*/([^/ ]*) [^/]*/([^\./]*)", l) # chain/resn seqid/atom.alt ; ignore alt
        if r:
            unk = r.group(2), r.group(1)

            cc = monlib.monomers[unk[1]] if unk[1] in monlib.monomers else None
            cc_atom = [x for x in cc.atoms if x.id == unk[0]] if cc else None
            if cc and cc_atom: # if atom is found in chemcomp, it must be an atom that should be removed.
                logger.write(l + " - this atom should have been removed when linking")
                if check_hydrogen or not cc_atom[0].is_hydrogen():
                    link_related.add(unk[1])
            elif unk[1] not in unknown_cc and unk not in unknown_atoms_cc:
                logger.write(l)
                unknown_atoms_cc.add(unk)
            continue

        # something else
        logger.write(l)

    if not check_hydrogen:
        todel = []
        for unk in unknown_atoms_cc:
            elements = [cra.atom.element for cra in st[0].all() if cra.residue.name==unk[1] and cra.atom.name==unk[0]]
            if elements and elements[0].is_hydrogen:
                todel.append(unk)
        unknown_atoms_cc = unknown_atoms_cc - set(todel)
        
    unknown_cc.update(cc for at, cc in unknown_atoms_cc)
        
    if stop_for_unknowns and (unknown_cc or link_related):
        msgs = []
        if unknown_cc: msgs.append("restraint cif file(s) for {}".format(",".join(unknown_cc)))
        if link_related: msgs.append("proper link cif file(s) for {} or check your model".format(",".join(link_related)))
        raise RuntimeError("Provide {}".format(" and ".join(msgs)))
   
    return monlib
# load_monomer_library()

def check_monlib_support_nucleus_distances(monlib, resnames):
    good = True
    nucl_not_found = []
    for resn in resnames:
        if resn not in monlib.monomers:
            logger.error("ERROR: monomer information of {} not loaded".format(resn))
            good = False
        else:
            mon = monlib.monomers[resn]
            no_nuc = False
            for bond in mon.rt.bonds:
                is_h = (mon.get_atom(bond.id1.atom).is_hydrogen(), mon.get_atom(bond.id2.atom).is_hydrogen())
                if any(is_h) and bond.value_nucleus != bond.value_nucleus:
                    no_nuc = True
                    break
            if no_nuc:
                nucl_not_found.append(resn)
                good = False

    if nucl_not_found:
        logger.write("WARNING: nucleus distance is not found for: {}".format(" ".join(nucl_not_found)))
        logger.write("         default scale ({}) is used for nucleus distances.".format(default_proton_scale))
    return good
# check_monlib_support_nucleus_distances()

def load_ener_lib(path=None):
    if path is None:
        if "CLIBD_MON" not in os.environ:
            logger.error("ERROR: CLIBD_MON is not set")
            return {}
        else:
            path = os.path.join(os.environ["CLIBD_MON"], "ener_lib.cif")

    if not os.path.exists(path):
        logger.error("ERROR: ener_lib file not found: {}".format(path))
        return {}

    enelib = gemmi.cif.read(path).sole_block()
    """
loop_
_lib_atom.type
_lib_atom.weight
_lib_atom.hb_type
_lib_atom.vdw_radius
_lib_atom.vdwh_radius
_lib_atom.ion_radius
_lib_atom.element
_lib_atom.valency
_lib_atom.sp
    """
    loop = enelib.find_loop("_lib_atom.type").get_loop()
    tags = [x.replace("_lib_atom.","") for x in loop.tags]
    ret = {}
    for i in range(loop.length()):
        key = loop.val(i, 0) # type
        #ret[key] = {}
        d = {}
        for j in range(1, loop.width()):
            val = loop.val(i, j)
            if tags[j] in ("weight", "vdw_radius", "vdwh_radius", "ion_radius"):
                d[tags[j]] = float(val) if val != "." else None
            elif tags[j] in ("valency", "sp"):
                d[tags[j]] = int(val) if val != "." else None
            else:
                d[tags[j]] = val

        ret[key] = d
    return ret
# load_ener_lib()

def find_and_fix_links(st, monlib, bond_margin=1.1, remove_unknown=False, add_found=True):
    """
    Find links not registered in st.connections
    This is required for correctly recognizing link in gemmi.prepare_topology
    if remove_unknown=True, undefined links and unmatched links are removed.
    """
    from servalcat.utils import model

    logger.write("Checking links in model")
    hunt = gemmi.LinkHunt()
    hunt.index_chem_links(monlib)
    matches = hunt.find_possible_links(st, bond_margin, 0)
    known_links = ("TRANS", "PTRANS", "NMTRANS", "CIS", "PCIS", "NMCIS", "p", "gap")
    conns = [x for x in st.connections] # to check later
    new_connections = []

    for m in matches:
        if m.conn:
            logger.write(" Link confirmed: {} atom1= {} atom2= {} dist= {:.2f} ideal= {:.2f}".format(m.chem_link.id,
                                                                                                    m.cra1, m.cra2,
                                                                                                    m.bond_length,
                                                                                                    m.chem_link.rt.bonds[0].value))
            if not m.cra1.atom_matches(m.conn.partner1): # need to swap
                assert m.cra1.atom_matches(m.conn.partner2)
                m.conn.partner1 = model.cra_to_atomaddress(m.cra1)
                m.conn.partner2 = model.cra_to_atomaddress(m.cra2)

            m.conn.link_id = m.chem_link.id
            if m.conn in conns: # may not be found if id duplicated
                conns.pop(conns.index(m.conn))
        else:
            # Known link is only accepted when in LINK record
            if not m.chem_link or m.chem_link.id in known_links:
                continue

            logger.write(" Link detected:  {} atom1= {} atom2= {} dist= {:.2f} ideal= {:.2f}".format(m.chem_link.id,
                                                                                                    m.cra1, m.cra2,
                                                                                                    m.bond_length,
                                                                                                    m.chem_link.rt.bonds[0].value))
            con = gemmi.Connection()
            con.type = gemmi.ConnectionType.Covale # XXX may be others
            con.link_id = m.chem_link.id
            con.partner1 = model.cra_to_atomaddress(m.cra1)
            con.partner2 = model.cra_to_atomaddress(m.cra2)
            new_connections.append(con)

    rm_idxes = []
    con_idxes = dict((c,i) for i,c in enumerate(st.connections))
    for con in conns:
        if con.link_id in known_links: continue
        if remove_unknown:
            i = con_idxes.get(con)
            if i is not None: rm_idxes.append(i)
            
        at1, at2 = st[0].find_cra(con.partner1).atom, st[0].find_cra(con.partner2).atom
        if None in (at1, at2):
            logger.write(" WARNING: atom(s) not found for link: atom1= {} atom2= {} id= {}".format(con.partner1, con.partner2, con.link_id))
            continue
        dist = at1.pos.dist(at2.pos)
        logger.write(" WARNING: unidentified link: atom1= {} atom2= {} dist= {:.2f} id= {}".format(con.partner1, con.partner2, dist, con.link_id))

    if remove_unknown:
        for i in sorted(rm_idxes, reverse=True):
            st.connections.pop(i)

    if add_found:
        for con in new_connections: # st.connections should have not been modified earlier because referenced in the loop above
            st.connections.append(con)

# find_and_fix_links()

def add_hydrogens(st, monlib, pos="elec"):
    assert pos in ("elec", "nucl")
    st.entities.clear()
    st.setup_entities()

    # Check links. XXX Is it ok to update st?
    find_and_fix_links(st, monlib, remove_unknown=True)
    
    topo = gemmi.prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.ReAddButWater, warnings=logger, ignore_unknown_links=True)
    if pos == "nucl":
        logger.write("Generating hydrogens at nucleus positions")
        resnames = st[0].get_all_residue_names()
        check_monlib_support_nucleus_distances(monlib, resnames)
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus, default_scale=default_proton_scale)
    else:
        logger.write("Generating hydrogens at electron positions")
# add_hydrogens()

def plane_deviations(atoms):
    pp = gemmi.find_best_plane(atoms)
    return [gemmi.get_distance_from_plane(a.pos, pp) for a in atoms]
# plane_deviations()

def nbc_z(nbc, nbc_ops):
    cra1,cra2,imgidx,_,mindist,sigma,_,_ = nbc
    
    if imgidx == 0:
        pos2 = cra2.atom.pos
    else:
        pos2 = gemmi.Position(nbc_ops[imgidx].apply(cra2.atom.pos))
        
    b = cra1.atom.pos.dist(pos2)
    db = b - mindist
    if db < 0:
        return db/sigma
    return 0.
# nbc_z()

class Restraints:
    #import line_profiler
    #profile = line_profiler.LineProfiler()
    #import atexit
    #atexit.register(profile.print_stats)
    #@profile
    def __init__(self, st, monlib, enerlib=None):
        self.st = st # clone?
        st.setup_cell_images()
        self.monlib = monlib
        self.enerlib = enerlib if enerlib is not None else load_ener_lib()
        find_and_fix_links(self.st, self.monlib)
        self.st.entities.clear()
        self.st.setup_entities() # entity information is needed for links
        self.topo = gemmi.prepare_topology(self.st, monlib, warnings=logger,
                                           ignore_unknown_links=True, reorder=True) # updates atom serial
        # lookup tables
        self.lookup = dict((x.atom, (i, x)) for i, x in enumerate(self.st[0].all()))
        self.nbc = []
        self.nbc_transforms = []
        self.outlier_sigmas = dict(bond=5, angle=5, torsion=5, nonbonded=5, chiral=5, plane=5)

        self.set_atom_types()
        self.construct_graph()
        self.find_nonbonded()
        self.set_ideal_chiral_volumes()
    # __init__()
    
    #@profile
    def construct_graph(self):
        import networkx as nx
        n_atoms = len(self.lookup)
        g = nx.Graph()
        for b in self.topo.bonds:
            idx1 = self.lookup[b.atoms[0]][0]
            idx2 = self.lookup[b.atoms[1]][0]
            g.add_edge(idx1, idx2)
            
        self.g = g
    # construct_graph()

    def set_atom_types(self):
        self.atom_types = {} # {resname: {atom_name: chem_type}}
        for resname, mon in self.monlib.monomers.items():
            self.atom_types[resname] = dict((a.id, a.chem_type) for a in mon.atoms)
    # set_atom_types()

    def get_chem_type(self, resname, atomname):
        r = self.atom_types.get(resname)
        if r: return r.get(atomname)
        return None
    # get_chem_type()

    #@profile
    def max_vdwr(self):
        maxr = 0
        for res in self.atom_types.values():
            for at in res.values():
                e = self.enerlib.get(at)
                if e:
                    maxr = max(maxr, e["vdw_radius"])
        logger.write("maximum vdw_radius in this model= {}".format(maxr))
        return maxr
    # max_vdwr()

    #@profile
    def get_nbc_mindist(self, type1, type2, d_12):
        default_min_dist = 3.4
        ene1 = self.enerlib.get(type1)
        ene2 = self.enerlib.get(type2)
        vdwtype = 1
        sigma = 0.2
        if None in (ene1, ene2):
            for t, e in ((type1, ene1), (type2, ene2)):
                if e is None: logger.write("WARNING: unknwon energy type: {}".format(t))
            return default_min_dist, sigma, vdwtype

        # Test hbond
        ht1, ht2 = ene1["hb_type"], ene2["hb_type"]
        vdwr1, vdwr2 = ene1["vdw_radius"], ene2["vdw_radius"]
        mindist = vdwr1 + vdwr2
        
        if ((ht1 in "AB" and ht2 in "DB") or
            (ht2 in "AB" and ht1 in "DB")):
            mindist -= 0.3 # ADHB
            sigma = 0.2
            vdwtype = 3

        if d_12 == 3: # 1_4 related atoms
            vdwtype = 2
            sigma = 0.2
            mindist -= 0.3

        # TODO hydrogen ignored!!
        # TODO in the same ring?
        return mindist, sigma, vdwtype

    # get_nbc_mindist()        

    #@profile
    def find_nonbonded(self):
        def check_bonds(s, e): # return shortest path between (s,e). checks up to 3 bonds
            if s not in self.g: return -1

            checked = {s: 0}
            for d in range(3):
                for n1 in [x for x in checked if checked[x]==d]:
                    for n2 in self.g.neighbors(n1):
                        if e == n2: return d+1
                        checked[n2] = d+1
            return -1
        # check_bonds()

        self.nbc = []
        self.nbc_transforms = []
        ns = gemmi.NeighborSearch(self.st[0], self.st.cell, 4).populate()
        cs = gemmi.ContactSearch(self.max_vdwr()*2)        
        results = cs.find_contacts(ns)
        for r in results: # TODO check alt loc!
            idx1 = self.lookup[r.partner1.atom][0]
            idx2 = self.lookup[r.partner2.atom][0]
            idx1_2_d = check_bonds(idx1, idx2)
            if idx1_2_d == 3 or idx1_2_d < 0:
                type1 = self.get_chem_type(r.partner1.residue.name, r.partner1.atom.name)
                type2 = self.get_chem_type(r.partner2.residue.name, r.partner2.atom.name)
                
                mindist, sigma, vdwtype = self.get_nbc_mindist(type1, type2, idx1_2_d)
                if r.dist < mindist:
                    #print('atom1="{}" atom2="{}" image={} dist={:.2f} id={:.2f} type={}'.format(r.partner1, r.partner2, r.image_idx, r.dist, mindist, vdwtype))
                    z = (mindist - r.dist)/sigma
                    self.nbc.append((r.partner1, r.partner2, r.image_idx, r.dist, mindist, sigma, z, vdwtype))

        # convert for orthogonal coordiinates
        transform = lambda x: gemmi.Transform(self.st.cell.orthogonalization_matrix.multiply(x.mat).multiply(self.st.cell.fractionalization_matrix),
                                              self.st.cell.orthogonalization_matrix.multiply(x.vec))
        if self.nbc:
            self.nbc_transforms = [transform(ns.get_image_transformation(i)) for i in range(max(x[2] for x in self.nbc)+1)]
    # find_nonbonded()

    #@profile
    def set_ideal_chiral_volumes(self):
        tmp = []
        for chir in self.topo.chirs:
            tmp.append(self.topo.ideal_chiral_abs_volume(chir))
            
        self.chir_vols = numpy.array(tmp)
    # set_ideal_chiral_volumes()

    #@profile
    def show_all(self):
        topo_d = dict(bond=self.topo.bonds, angle=self.topo.angles, torsion=self.topo.torsions,
                      chiral=self.topo.chirs, plane=self.topo.planes, nonbonded=self.nbc)
        dfs = {}
        for k in topo_d:
            topo_k = topo_d[k]

            if k == "torsion":
                atom_labels = ["atom{}".format(i+1) for i in range(4)]
                df = pandas.DataFrame([[str(self.lookup[y][1]) for y in x.atoms] + [x.calculate(), x.restr.value,
                                                                            x.restr.period, x.restr.esd, x.calculate_z(), x.restr.label]
                                       for x in topo_k],
                                      columns= atom_labels + ["model", "ideal", "period", "esd", "z", "label"])
                df = df.loc[df.esd > 0]
            elif k == "chiral":
                idealstr = {gemmi.ChiralityType.Positive:"positive",
                            gemmi.ChiralityType.Negative:"negative",
                            gemmi.ChiralityType.Both:"both"}
                df = pandas.DataFrame([[str(self.lookup[y][1]) for y in x.atoms] + [x.calculate(), self.topo.ideal_chiral_abs_volume(x),
                                                                            idealstr[x.restr.sign], 0.2, x.check()] for x in topo_k],
                                      columns=["atomc", "atom1", "atom2", "atom3", "model", "ideal_abs", "sign", "esd", "check"])
                df["z"] = numpy.abs([b.calculate_z(iv, 0.2) for b, iv in zip(topo_k, df["ideal_abs"])])
            elif k == "plane":
                index = pandas.MultiIndex.from_arrays([["{}_{}".format(i, x.restr.label) for i, x in enumerate(topo_k) for y in x.atoms],
                                                       [str(self.lookup[y][1]) for x in topo_k for y in x.atoms]],
                                                       names=["label", "atom"])
                df = pandas.DataFrame([(d, x.restr.esd) for i,x in enumerate(topo_k)
                                       for y,d in zip(x.atoms, plane_deviations(x.atoms))],
                                      index=index,
                                      columns=["dev", "esd"])
                df["z"] = numpy.abs(df.dev/df.esd)
            elif k == "nonbonded":
                df = pandas.DataFrame([(str(partner1), str(partner2), image_idx, dist, mindist, vdwtype, sigma, z)
                                       for partner1, partner2, image_idx, dist, mindist, sigma, z, vdwtype in topo_k],
                                      columns=["atom1", "atom2", "image", "model", "mindist", "vdwtype", "esd", "z"])
            else:
                n_atoms = dict(bond=2, angle=3, torsion=4)
                atom_labels = ["atom{}".format(i+1) for i in range(n_atoms[k])]
                df = pandas.DataFrame([[str(self.lookup[y][1]) for y in x.atoms] + [x.calculate(), x.restr.value, x.restr.esd, x.calculate_z()]
                                       for x in topo_k],
                                      columns= atom_labels + ["model", "ideal", "esd", "z"])

            if k in ("angle", "torsion"): df["model"] = numpy.rad2deg(df["model"])
            dfs[k] = df

            if k == "chiral":
                df = df.loc[(~df.check) | (df.z > self.outlier_sigmas[k])].sort_values("z", ascending=False)
            elif k == "plane":
                # sort by z within label (same plane)
                df = df.reset_index().sort_values(["label", "z"], ascending=[True, False]).set_index(["label", "atom"])
                # sort by max z
                tmp = df.groupby(level=0)["z"].max()
                idx=tmp[tmp > self.outlier_sigmas[k]].sort_values(ascending=False).index
                df = df.loc[idx]
            else:
                df = df.loc[(df.z > self.outlier_sigmas[k])].sort_values("z", ascending=False)

            n_outl = len(df.index.unique(0)) if k == "plane" else len(df.index)
            logger.write("\n{} {} outliers (> {} sigma)".format(n_outl, k, self.outlier_sigmas[k]))
            if n_outl > 0:
                logger.write(df.to_string(index=(k=="plane")))

        logger.write("\nSummary:")
        tmp = []
        for k in dfs:
            df = dfs[k]
            if k == "nonbonded": df = df.loc[df.model < df.mindist]
            tmp.append([k, len(df.index), numpy.sqrt(numpy.mean((df.z*df.esd)**2)), numpy.mean(df.esd), numpy.sqrt(numpy.mean(df.z**2))])
            if k == "torsion":
                for p, g in df.groupby("period", sort=True):
                    tmp.append(["{} (period {})".format(k, p), len(g.index), numpy.sqrt(numpy.mean((g.z*g.esd)**2)),
                                numpy.mean(g.esd), numpy.sqrt(numpy.mean(g.z**2))])
            elif k == "chiral":
                df = df.loc[df.sign!="both"]
                tmp.append(["{} (non-both)".format(k), len(df.index), numpy.sqrt(numpy.mean((df.z*df.esd)**2)),
                            numpy.mean(df.esd), numpy.sqrt(numpy.mean(df.z**2))])

            
        logger.write(pandas.DataFrame(tmp,
                                      columns=["Restraint type", "number", "rmsd", "mean(esd)", "rmsz"],
                                      ).to_string(index=False, float_format="%.3f"))

        logger.write("\n!WARNING!: this function has problems at the moment. Will be sorted out later.")
        logger.write("""\
TODO
- nbc distance for metals (e.g. Mg-O) and hydrogen atoms are not correct
- redundant restraint definitions should be sorted out (e.g. A has C2e-nyu0 and C3e-nyu0 for the same atoms)
- separate refined and non-refined atoms (hydrogen, occ 0?)""")

        return dfs
    # show_all()
# class Restraints



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import os
import shutil
import argparse
from servalcat.utils import logger
from servalcat import utils
from servalcat import spa

def add_arguments(parser):
    parser.description = 'Run REFMAC5 for SPA'

    parser.add_argument('--exe', default="refmac5",
                        help='refmac5 binary (default: %(default)s)')
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    # sfcalc options
    sfcalc_group = parser.add_argument_group("sfcalc")
    spa.sfcalc.add_sfcalc_args(sfcalc_group)

    # run_refmac options
    # TODO use group! like refmac options
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--bfactor', type=float,
                        help="reset all atomic B values to specified value")
    parser.add_argument('--ncsr', default="local", choices=["local", "global"],
                        help="local or global NCS restrained (default: %(default)s)")
    parser.add_argument('--ncycle', type=int, default=10,
                        help="number of cycles in Refmac (default: %(default)d)")
    parser.add_argument('--tlscycle', type=int, default=0,
                        help="number of TLS cycles in Refmac (default: %(default)d)")
    parser.add_argument('--tlsin',
                        help="TLS parameter input for Refmac")
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input. "
                        "Default: %(default)s")
    parser.add_argument('--jellybody', action='store_true',
                        help="Use jelly body restraints")
    parser.add_argument('--jellybody_params', nargs=2, type=float,
                        metavar=("sigma", "dmax"), default=[0.01, 4.2],
                        help="Jelly body sigma and dmax (default: %(default)s)")
    parser.add_argument('--hout', action='store_true', help="write hydrogen atoms in the output model")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--weight_auto_scale', type=float,
                       help="'weight auto' scale value. automatically determined from resolution and mask/box volume ratio if unspecified")
    group.add_argument('--weight_matrix', type=float,
                       help="weight matrix value")
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('--external_restraints_json')
    parser.add_argument('--show_refmac_log', action='store_true',
                        help="show all Refmac log instead of summary")
    parser.add_argument('--output_prefix', default="refined",
                        help='output file name prefix (default: %(default)s)')
    parser.add_argument('--cross_validation', action='store_true',
                        help='Run cross validation')
    parser.add_argument('--cross_validation_method', default="shake", choices=["throughout", "shake"],
                        help="shake: randomize a model refined against a full map and then refine it against a half map, "
                        "throughout: use only a half map for refinement (another half map is used for error estimation) "
                        "Default: %(default)s")
    parser.add_argument('--shake_radius', default=0.3,
                        help='Shake rmsd in case of --cross_validation_method=shake (default: %(default).1f)')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--mask_for_fofc', help="Mask file for Fo-Fc map calculation")
    group.add_argument('--mask_radius_for_fofc', type=float, help="Mask radius for Fo-Fc map calculation")
    parser.add_argument('--trim_fofc_mtz', action="store_true", help="diffmap.mtz will have smaller cell (if --mask_for_fofc is given)")
    parser.add_argument("--fsc_resolution", type=float,
                        help="High resolution limit for FSC calculation. Default: Nyquist")

# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_fsc(st, output_prefix, maps, d_min, mask, mask_radius, b_before_mask, no_sharpen_before_mask, make_hydrogen, monlib,
             blur=None, d_min_fsc=None, cross_validation=False, cross_validation_method=None, st_sr=None):
    # st_sr: shaken-and-refined st in case of cross_validation_method=="shake"
    if cross_validation:
        assert len(maps) == 2
        assert cross_validation_method in ("shake", "throughout")
    if cross_validation and cross_validation_method == "shake":
        assert st_sr is not None
    else:
        assert st_sr is None
    
    logger.write("Calculating map-model FSC..")

    if d_min_fsc is None:
        d_min_fsc = utils.maps.nyquist_resolution(maps[0][0])
        logger.write("  --fsc_resolution is not specified. Using Nyquist resolution: {:.2f}".format(d_min_fsc))
    
    st = st.clone()
    if st_sr is not None: st_sr = st_sr.clone()

    if make_hydrogen == "all":
        utils.restraints.add_hydrogens(st, monlib)
        if st_sr is not None: utils.restraints.add_hydrogens(st_sr, monlib)

    if mask is not None or mask_radius is not None:
        if mask is None:
            mask = gemmi.FloatGrid(*maps[0][0].shape)
            mask.set_unit_cell(st.cell)
            mask.spacegroup = st.find_spacegroup()
            mask.mask_points_in_constant_radius(st[0], mask_radius, 1.)
        if no_sharpen_before_mask or len(maps) < 2:
            for ma in maps: ma[0].array[:] *= mask
        else:
            # It seems we need different B for different resolution limit
            if b_before_mask is None: b_before_mask = spa.sfcalc.determine_b_before_mask(st, maps, maps[0][1], mask, d_min_fsc)
            maps = utils.maps.sharpen_mask_unsharpen(maps, mask, d_min_fsc, b=b_before_mask)
        
    hkldata = utils.maps.mask_and_fft_maps(maps, d_min_fsc)
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, d_min_fsc - 1e-6, cutoff=1e-7, monlib=monlib, source="electron",
                                               miller_array=hkldata.miller_array())
    labs_fc = ["FC"]

    if st_sr is not None:
        hkldata.df["FC_sr"] = utils.model.calc_fc_fft(st_sr, d_min_fsc - 1e-6, cutoff=1e-7, monlib=monlib, source="electron",
                                                      miller_array=hkldata.miller_array())
        labs_fc.append("FC_sr")

    if blur is not None:
        logger.write(" Unblurring Fc with B={} for FSC calculation".format(blur))
        unblur = numpy.exp(blur/hkldata.d_spacings().to_numpy()**2/4.)
        for lab in labs_fc:
            hkldata.df[lab] *= unblur
    
    hkldata.setup_relion_binning()
    stats = spa.fsc.calc_fsc(hkldata, labs_fc=labs_fc, lab_f="FP",
                             labs_half=["F_map1", "F_map2"] if len(maps)==2 else None)

    hkldata2 = hkldata.copy(d_min=d_min) # for FSCaverage at resolution for refinement # XXX more efficient way
    hkldata2.setup_relion_binning()
    stats2 = spa.fsc.calc_fsc(hkldata2, labs_fc=labs_fc, lab_f="FP",
                              labs_half=["F_map1", "F_map2"] if len(maps)==2 else None)
    
    if "fsc_half" in stats:
        with numpy.errstate(invalid="ignore"): # XXX negative fsc results in nan!
            stats.loc[:,"fsc_full_sqrt"] = numpy.sqrt(2*stats.fsc_half/(1+stats.fsc_half))

    logger.write(stats.to_string()+"\n")

    # remove and rename columns
    for s in (stats, stats2):
        s.rename(columns=dict(fsc_FC_full="fsc_model", Rcmplx_FC_full="Rcmplx"), inplace=True)
        if cross_validation:
            if cross_validation_method == "shake":
                s.drop(columns=["fsc_FC_half1", "fsc_FC_half2", "fsc_FC_sr_full", "Rcmplx_FC_sr_full"], inplace=True)
                s.rename(columns=dict(fsc_FC_sr_half1="fsc_model_half1",
                                          fsc_FC_sr_half2="fsc_model_half2"), inplace=True)
            else:
                s.rename(columns=dict(fsc_FC_half1="fsc_model_half1",
                                          fsc_FC_half2="fsc_model_half2"), inplace=True)
        else:
            s.drop(columns=[x for x in s if x.startswith("fsc_FC") and x.endswith(("half1","half2"))], inplace=True)

    # FSCaverages
    fscavg_text  = "Map-model FSCaverages (at {:.2f} A):\n".format(d_min)
    fscavg_text += " FSCaverage(full) = {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model))
    if cross_validation:
        fscavg_text += "Cross-validated map-model FSCaverages:\n"
        fscavg_text += " FSCaverage(half1)= {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model_half1))
        fscavg_text += " FSCaverage(half2)= {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model_half2))
    
    # for loggraph
    fsc_logfile = "{}_fsc.log".format(output_prefix)
    with open(fsc_logfile, "w") as ofs:
        columns = "1/resol^2 ncoef ln(Mn(|Fo|^2)) ln(Mn(|Fc|^2)) FSC(full,model) FSC_half FSC_full_sqrt FSC(half1,model) FSC(half2,model) Rcmplx(full,model)".split()
        
        ofs.write("$TABLE: Map-model FSC after refinement:\n")
        if len(maps) == 2:
            if cross_validation: fsc_cols = [5,6,7,8,9]
            else:                fsc_cols = [5,6,7]
        else: fsc_cols = [5]
        fsc_cols.append(10)
        if len(maps) == 2: ofs.write("$GRAPHS: FSC :A:1,5,6,7:\n")
        else:              ofs.write("$GRAPHS: FSC :A:1,5:\n")
        if cross_validation: ofs.write(": cross-validated FSC :A:1,8,9:\n".format(",".join(map(str,fsc_cols))))
        ofs.write(": Rcmplx :A:1,{}:\n".format(4+len(fsc_cols)))
        ofs.write(": ln(Mn(|F|^2)) :A:1,3,4:\n")
        ofs.write(": number of Fourier coeffs :A:1,2:\n")
        ofs.write("$$ {}$$\n".format(" ".join(columns[:4]+[columns[i-1] for i in fsc_cols])))
        ofs.write("$$\n")

        plot_columns = ["d_min", "ncoeffs", "power_FP", "power_FC", "fsc_model"]
        if len(maps) == 2:
            plot_columns.extend(["fsc_half", "fsc_full_sqrt"])
            if cross_validation:
                plot_columns.extend(["fsc_model_half1", "fsc_model_half2"])
        plot_columns.append("Rcmplx")
        with numpy.errstate(divide="ignore"):
            log_format = lambda x: "{:.3f}".format(numpy.log(x))
            ofs.write(stats.to_string(header=False, index=False, index_names=False, columns=plot_columns,
                                      formatters=dict(d_min=lambda x: "{:.4f}".format(1/x**2),
                                                      power_FP=log_format, power_FC=log_format)))
        ofs.write("\n")
        ofs.write("$$\n\n")
        ofs.write(fscavg_text)
    
    logger.write(fscavg_text, end="")
    logger.write("Run loggraph {} to see plots.".format(fsc_logfile))
    
    # write json
    json.dump(stats.to_dict("records"),
              open("{}_fsc.json".format(output_prefix), "w"),
              indent=True)

    return fscavg_text
# calc_fsc()

def main(args):
    if not args.model:
        raise SystemExit("Error: give --model.")

    if not (args.map or args.halfmaps):
        raise SystemExit("Error: give --map | --halfmaps.")

    if args.mask_for_fofc and not os.path.exists(args.mask_for_fofc):
        raise SystemExit("Error: --mask_for_fofc {} does not exist".format(args.mask_for_fofc))

    if args.mask_for_fofc and args.mask_radius_for_fofc:
        raise SystemExit("Error: you cannot specify both --mask_for_fofc and --mask_radius_for_fofc")

    if args.trim_fofc_mtz and not (args.mask_for_fofc or args.mask_radius_for_fofc):
        raise SystemExit("Error: --trim_fofc_mtz is specified but --mask_for_fofc is not given")

    if args.ligand: args.ligand = sum(args.ligand, [])

    st = utils.fileio.read_structure(args.model)
    try:
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand, 
                                                       stop_for_unknowns=True, check_hydrogen=(args.hydrogen=="yes"))
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    if args.mask_for_fofc and not args.no_check_mask_with_model:
        mask = utils.fileio.read_ccp4_map(args.mask_for_fofc)[0]
        if not utils.maps.test_mask_with_model(mask, st):
            raise SystemExit("\nError: Model is out of mask.\n"
                             "Please check your --model and --mask_for_fofc. You can disable this test with --no_check_mask_with_model.")
        
    args.shifted_model_prefix = "shifted"
    args.output_masked_prefix = "masked_fs"
    args.output_mtz_prefix = "starting_map"
    file_info = spa.sfcalc.main(args, monlib=monlib)
    args.mtz = file_info["mtz_file"]
    if args.halfmaps: # FIXME if no_mask?
        args.mtz_half = [file_info["mtz_file"], file_info["mtz_file"]]
    args.lab_phi = file_info["lab_phi"]  #"Pout0"
    args.lab_f = file_info["lab_f"]
    args.lab_sigf = None
    args.model = file_info["model_file"]
    model_format = file_info["model_format"]

    if args.cross_validation and args.cross_validation_method == "throughout":
        args.lab_f = file_info["lab_f_half1"]
        args.lab_phi = file_info["lab_phi_half1"]
        # XXX args.lab_sigf?

    if args.keyword_file:
        args.keyword_file = sum(args.keyword_file, [])
        for f in args.keyword_file:
            logger.write("Keyword file: {}".format(f))
            assert os.path.exists(f)
    else:
        args.keyword_file = []
            
    if args.keywords:
        args.keywords = sum(args.keywords, [])

    chain_id_lens = [len(x) for x in utils.model.all_chain_ids(st)]
    keep_chain_ids = (chain_id_lens and max(chain_id_lens) == 1) # always kept unless one-letter chain IDs

    # FIXME if mtz is given and sfcalc() not ran?
    has_ncsc = "ncsc_file" in file_info
    if has_ncsc:
        args.keyword_file.append(file_info["ncsc_file"])

    if not args.no_trim:
        refmac_prefix = "{}_{}".format(args.shifted_model_prefix, args.output_prefix)
    else:
        refmac_prefix = args.output_prefix # XXX this should be different name (nomask etc?)

    # Weight auto scale
    if args.weight_matrix is None and args.weight_auto_scale is None:
        reso = file_info["d_eff"] if "d_eff" in file_info else args.resolution
        if "vol_ratio" in file_info:
            if "d_eff" in file_info:
                rlmc = (-9.503541, 3.129882, 15.439744)
            else:
                rlmc = (-8.329418, 3.032409, 14.381907)
            logger.write("Estimating weight auto scale using resolution and volume ratio")
            ws = rlmc[0] + reso*rlmc[1] +file_info["vol_ratio"]*rlmc[2]
        else:
            if "d_eff" in file_info:
                rlmc = (-5.903807, 2.870723)
            else:
                rlmc = (-4.891140, 2.746791)
            logger.write("Estimating weight auto scale using resolution")
            ws =  rlmc[0] + args.resolution*rlmc[1]
        args.weight_auto_scale = max(0.2, min(18.0, ws))
        logger.write(" Will use weight auto {:.2f}".format(args.weight_auto_scale))

    # Take care of TLS in
    if args.tlsin and not args.no_shift and "shifts" in file_info:
        logger.write("Shifting origin in tlsin")
        tlsgroups = utils.refmac.read_tls_file(args.tlsin)
        spa.shiftback.shift_back_tls(tlsgroups, -file_info["shifts"]) # tlsgroups is modified
        args.tlsin = "shifted.tls"
        utils.refmac.write_tls_file(tlsgroups, args.tlsin)
        
    # Run Refmac
    refmac = utils.refmac.Refmac(prefix=refmac_prefix, args=args, global_mode="spa",
                                 keep_chain_ids=keep_chain_ids)
    refmac.set_libin(args.ligand)
    try:
        refmac_summary = refmac.run_refmac()
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    json.dump(refmac_summary,
              open("{}_summary.json".format(refmac_prefix), "w"),
              indent=True)

    if args.halfmaps:
        maps = [utils.fileio.read_ccp4_map(f, pixel_size=args.pixel_size) for f in args.halfmaps]
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]

    # Modify output
    st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(refmac_prefix+model_format)
    utils.model.adp_analysis(st)
    
    if not args.no_trim:
        st.cell = maps[0][0].unit_cell
        if not args.no_shift:
            spa.shiftback.shift_back_model(st, file_info["shifts"]) # st is modified
    
    if "refmac_fixes" in file_info:
        file_info["refmac_fixes"].modify_back(st)
    utils.fileio.write_model(st, prefix=args.output_prefix,
                             pdb=True, cif=True, cif_ref=cif_ref)

    # Take care of TLS out
    if not args.no_trim: # if no_trim, there is nothing to do.
        tlsout = refmac.tlsout()
        if os.path.exists(tlsout):
            if not args.no_shift:
                logger.write("Shifting origin in tlsout")
                tlsgroups = utils.refmac.read_tls_file(tlsout)
                spa.shiftback.shift_back_tls(tlsgroups, file_info["shifts"]) # tlsgroups is modified
                utils.refmac.write_tls_file(tlsgroups, args.output_prefix+".tls")
            else:
                logger.write("Copying tlsout")
                shutil.copyfile(refmac.tlsout(), args.output_prefix+".tls")

    # Expand sym here
    st_expanded = st.clone()
    if has_ncsc:
        utils.model.expand_ncs(st_expanded)
        utils.fileio.write_model(st_expanded, file_name=args.output_prefix+"_expanded"+model_format,
                                 cif_ref=cif_ref)

    if args.cross_validation and args.cross_validation_method == "shake":
        logger.write("Cross validation is requested.")
        refmac_prefix_shaken = refmac_prefix+"_shaken_refined"
        logger.write("Starting refinement using half map 1 (model is shaken first)")
        refmac_hm1 = refmac.copy(hklin=args.mtz_half[0],
                                 xyzin=refmac_prefix+model_format,
                                 prefix=refmac_prefix_shaken,
                                 shake=args.shake_radius,
                                 jellybody=False) # makes no sense to use jelly body after shaking
        if args.jellybody: logger.write("  Turning off jelly body")
        if "lab_f_half1" in file_info:
            refmac_hm1.lab_f = file_info["lab_f_half1"]
            refmac_hm1.lab_phi = file_info["lab_phi_half1"]
            # SIGMA?
            
        try:
            refmac_hm1.run_refmac()
        except RuntimeError as e:
            raise SystemExit("Error: {}".format(e))
        
        # Modify output
        st_sr, cif_ref_sr = utils.fileio.read_structure_from_pdb_and_mmcif(refmac_prefix_shaken+model_format)
        if not args.no_trim:
            st_sr.cell = maps[0][0].unit_cell
            if not args.no_shift:
                spa.shiftback.shift_back_model(st_sr, file_info["shifts"])
            
        if "refmac_fixes" in file_info:
            file_info["refmac_fixes"].modify_back(st)

        utils.fileio.write_model(st_sr, prefix=args.output_prefix+"_shaken_refined",
                                 pdb=True, cif=True, cif_ref=cif_ref_sr)

        # Expand sym here
        st_sr_expanded = st_sr.clone()
        if has_ncsc:
            utils.model.expand_ncs(st_sr_expanded)
            utils.fileio.write_model(st_sr_expanded, file_name=args.output_prefix+"shaken_refined_expanded"+model_format,
                                     cif_ref=cif_ref_sr)

    else:
        st_sr_expanded = None

    if args.mask:
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None
        
    # Calc FSC
    fscavg_text = calc_fsc(st_expanded, args.output_prefix, maps,
                           args.resolution, mask=mask, mask_radius=args.mask_radius if not args.no_mask else None,
                           b_before_mask=args.b_before_mask,
                           no_sharpen_before_mask=args.no_sharpen_before_mask,
                           make_hydrogen=args.hydrogen,
                           monlib=monlib, cross_validation=args.cross_validation,
                           blur=args.blur[0] if args.blur else None,
                           d_min_fsc=args.fsc_resolution,
                           cross_validation_method=args.cross_validation_method, st_sr=st_sr_expanded)

    # Calc Fo-Fc (and updated) maps
    logger.write("Starting Fo-Fc calculation..")
    if not args.halfmaps: logger.write(" with limited functionality because half maps were not provided")
    logger.write(" model: {}".format(args.output_prefix+model_format))

    # for Fo-Fc in case of helical reconstruction, expand model more
    # XXX should we do it for FSC calculation also? Probably we should not do sharpen-unsharpen procedure for FSC calc either.
    if args.twist is not None:
        logger.write("Generating all helical copies in the box")
        st_expanded = st.clone()
        utils.symmetry.update_ncs_from_args(args, st_expanded, map_and_start=maps[0], filter_model_helical_contacting=False)
        utils.model.expand_ncs(st_expanded)
        utils.fileio.write_model(st_expanded, file_name=args.output_prefix+"_expanded_all"+model_format,
                                 cif_ref=cif_ref)

    if args.mask_for_fofc:
        logger.write("  mask: {}".format(args.mask_for_fofc))
        mask = utils.fileio.read_ccp4_map(args.mask_for_fofc)[0]
    elif args.mask_radius_for_fofc:
        logger.write("  mask: using refined model with radius of {} A".format(args.mask_radius_for_fofc))
        mask = gemmi.FloatGrid(*maps[0][0].shape)
        mask.set_unit_cell(maps[0][0].unit_cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        mask.mask_points_in_constant_radius(st_expanded[0], args.mask_radius_for_fofc, 1.)
    else:
        logger.write("  mask: not used")
        mask = None
        
    hkldata, map_labs, stats_str = spa.fofc.calc_fofc(st_expanded, args.resolution, maps, mask=mask, monlib=monlib,
                                                      half1_only=(args.cross_validation and args.cross_validation_method == "throughout"),
                                                      sharpening_b=None if args.halfmaps else 0.) # assume already sharpened if fullmap is given
    spa.fofc.write_files(hkldata, map_labs, maps[0][1], stats_str,
                         mask=mask, output_prefix="diffmap",
                         trim_map=mask is not None, trim_mtz=args.trim_fofc_mtz)

    # Final summary
    if len(refmac_summary["cycles"]) > 1 and "actual_weight" in refmac_summary["cycles"][-2]:
        final_weight = refmac_summary["cycles"][-2]["actual_weight"]
    else:
        final_weight = "???"

    adpstats_txt = ""
    adp_stats = utils.model.adp_stats_per_chain(st[0])
    max_chain_len = max([len(x[0]) for x in adp_stats])
    max_num_len = max([len(str(x[1])) for x in adp_stats])
    for chain, natoms, qs in adp_stats:
        adpstats_txt += " Chain {0:{1}s}".format(chain, max_chain_len) if chain!="*" else " {0:{1}s}".format("All", max_chain_len+6)
        adpstats_txt += " ({0:{1}d} atoms) min={2:5.1f} median={3:5.1f} max={4:5.1f} A^2\n".format(natoms, max_num_len, qs[0],qs[2],qs[4])

    # Create Coot script
    with open("{}_coot.py".format(args.output_prefix), "w") as ofs:
        ofs.write('imol = read_pdb("{}.pdb")\n'.format(args.output_prefix)) # as Coot is not good at mmcif file..
        ofs.write('imol_fo = make_and_draw_map("diffmap.mtz", "FWT", "PHWT", "", 0, 0)\n')
        ofs.write('imol_fofc = make_and_draw_map("diffmap.mtz", "DELFWT", "PHDELWT", "", 0, 1)\n')
        if mask is not None:
            ofs.write('set_contour_level_absolute(imol_fo, 1.2)\n')
            ofs.write('set_contour_level_absolute(imol_fofc, 3.0)\n')
        
    logger.write("""
=============================================================================
* Final Summary *

Rmsd from ideal
  bond lengths: {rmsbond} A
  bond  angles: {rmsangle} deg

{fscavgs}
 Run loggraph {fsclog} to see plots

ADP statistics
{adpstats}

Weight used: {final_weight}
             If you want to change the weight, give larger (looser restraints)
             or smaller (tighter) value to --weight_auto_scale=.
             
Open refined model and diffmap.mtz with COOT:
coot --script {prefix}_coot.py
=============================================================================
""".format(rmsbond=refmac_summary["cycles"][-1].get("rms_bond", "???"),
           rmsangle=refmac_summary["cycles"][-1].get("rms_angle", "???"),
           fscavgs=fscavg_text.rstrip(),
           fsclog="{}_fsc.log".format(args.output_prefix),
           adpstats=adpstats_txt.rstrip(),
           final_weight=final_weight,
           prefix=args.output_prefix))
# main()
        
if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)

